#!/usr/bin/perl

# Copyright (c) 2009, Aldis Berjoza <killasmurf86@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above
#   copyright notice, this list of conditions and the following disclaimer
#   in the documentation and/or other materials provided with the
#   distribution.
# * Neither the name of the  nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# It's ugly down there :D

###################################
# DEPENDENCIES
###################################
# perl
# devel/p5-File-Type
# multimedia/mplayer
###################################

package playd;

use warnings;
use strict;

use Cwd qw(abs_path getcwd);
use Errno qw(EAGAIN);
use File::Type;
use POSIX qw(mkfifo);
use File::Spec::Functions qw(rootdir);
use List::Util qw(shuffle);

my $v_major = 2;
my $v_minor = 1;
my $v_update = 0;
my $v_devel = 0;
my $playd_version = "$v_major.$v_minor.$v_update";
$playd_version .= "_$v_devel" if $v_devel;

(my $playd_name = $0) =~ s/.*?\///g; # what's my name?

# some common variables
my $playd_home = "$ENV{HOME}/.config/playd2";
my $playd_lock = "$playd_home/lock";
my $playd_fifo = "$playd_home/fifo";
my $playd_playlist = "$playd_home/playlist.plst";
my $playd_tmp_playlist =  "$playd_home/tmp_playlist.plst";
my $playd_mplayer_cmd = "mplayer -idle -input file=$playd_fifo";
my $playd_mplayer_no_video_cmd = "mplayer -vo /dev/null -idle -input file=$playd_fifo";
my $playd_append = 0;
my $playd_novid = 0;
my $playd_console = 0;
my $playd_mplayer_user_options = '';

my $ft = File::Type->new();

unless ( -d $playd_home ) {
	mkdir($playd_home, 0700) or die "ERR: can't create directory '$playd_home'.\n$!\n";
}

unless ( -p $playd_fifo ) {
	mkfifo($playd_fifo, 0600) or die "ERR: can't create fifo '$playd_home'.\n$!\n";
}

&get_args() if ( defined $ARGV[0] );
&help();
exit;

#########################################################################################
#########################################################################################
#########################################################################################

sub get_args {
	if ( $ARGV[0] =~ /^((--)??append|-a)$/ ) {
		shift @ARGV;
		$playd_append = 1;
	}

	while ( $ARGV[0] ) {

# --append
		if ( $ARGV[0] =~ /^((--)??append|-a)$/ ) {
			warn "WARN: '$ARGV[0]' should be first argument on command line. Ignoring\n";
		}

# --cd
# --dvd
		elsif ( $ARGV[0] =~ /^((--)??(cd(da)??|dvd)|-c|-d)$/ ) {
			my $media;
			$media = 'cdda' if ( $ARGV[0] =~ /^((--)??cd(da)??|-c)$/ );
			$media = 'dvd' if ( $ARGV[0] =~ /^((--)??dvd|-d)$/ );

			if ( defined $ARGV[1] and $ARGV[1] =~ /^\d$/ ) {
				while ( $ARGV[1] =~ /^\d$/ ) {
					&add_to_playlist("$media://$ARGV[1]");
					shift @ARGV;
				}
			}
			else {
				&add_to_playlist("$media://");
			}
		}

# --help
		elsif ( $ARGV[0] =~ /^((--)??help|-h)$/ ) {
			&help;
		}

# --list
		elsif ( $ARGV[0] =~ /^((--)??list|-l)$/ ) {
			&list(0);
		}

# --longlist
		elsif ( $ARGV[0] =~ /^((--)??l(ong)??list|-L)$/ ) {
			&list(1);
		}

# --mute
		elsif ( $ARGV[0] =~ /^((--)??mute|-m)$/ ) {
			&send_mplayer_cmd('mute');
		}

# --playlist
		elsif ( $ARGV[0] =~ /^((--)??playlist|-P)$/ ) {
			if ( defined $ARGV[1] and -f $ARGV[1] and $ARGV[1] =~ /\.plst^/ ) {
				&start;
				&send_mplayer_cmd("loadlist $ARGV[1] $playd_append");
				$playd_append = 1;
				shift @ARGV;
			}
			elsif ( -f $playd_playlist ) {
				&start;
				&send_mplayer_cmd("loadlist '$playd_playlist' $playd_append");
				$playd_append = 1;
			}
			else {
				warn "WARN: default playlist is not available. Ignoring\n";
			}
		}

# --play
		elsif ( $ARGV[0] =~ /^((--)??play|-p)$/ ) {
			if ( defined $ARGV[1] and $ARGV[1] =~ /^\d+$/ ) {
				&start;
				open(PLAYLIST, '<', $playd_playlist) or die "ERR: can't open file '$playd_playlist'.\n$!\n";
				my @item_pos;
				push(@item_pos, 0);
				while ( my $item = <PLAYLIST> ) {
					push(@item_pos, tell(PLAYLIST));
				}
				
				my @tmp_list;
				while ( defined $ARGV[1] and $ARGV[1] =~ /^\d+$/ ) {
					if ( defined $item_pos[ $ARGV[1] - 1 ] ) {
						seek(PLAYLIST, $item_pos[$ARGV[1] - 1], 0);
						push(@tmp_list, my $tmp_item = <PLAYLIST>);
					}
					else {
						warn "WARN: there's no playlist item with number '$ARGV[1]'. Ignoring\n";
					}
					shift @ARGV;
				}
				close PLAYLIST;

				open (TMP_PLAYLIST, '>', $playd_tmp_playlist) or die "ERR: can\'t open file '$playd_tmp_playlist' for writing.\n$!\n";
				print TMP_PLAYLIST @tmp_list;
				close TMP_PLAYLIST;
				&send_mplayer_cmd("loadlist '$playd_tmp_playlist' $playd_append");
				$playd_append = 1;
				sleep 1;
				unlink $playd_tmp_playlist;

			}
		}

# --randomize
		elsif ( $ARGV[0] =~ /^(--)??(randomize|rnd)$/ ) {
			open(PLAYLIST, '<', $playd_playlist) or die "ERR: can't open file '$playd_playlist'.\n$!\n";
			my @item_pos;
			push(@item_pos, 0);
			while ( my $item = <PLAYLIST> ) {
				push(@item_pos, tell(PLAYLIST));
			}
			pop @item_pos, my $playlist_eof; # I don't need to know that
			
			open (TMP_PLAYLIST, '>', $playd_tmp_playlist) or die "ERR: can\'t open file '$playd_tmp_playlist' for writing.\n$!\n";
			@item_pos = shuffle(@item_pos);
			foreach my $item (@item_pos) {
				seek(PLAYLIST, $item, 0);
				my $entry = <PLAYLIST>;
				print TMP_PLAYLIST $entry;
			}

			close PLAYLIST;
			close TMP_PLAYLIST;
			unlink $playd_playlist;
			rename $playd_tmp_playlist, $playd_playlist;
		}

# --quit
		elsif ( $ARGV[0] =~ /^((--)??(quit|stop)|-q)$/ ) {
			&quit;
		}

# --status
		elsif ( $ARGV[0] =~ /^((--)??status|-i)$/ ) {
			my $pid = &get_mplayer_pid;
			if ( $pid ) {
				print "playd is running. PID: $pid\n";
			}
			else {
				print "playd is not running\n";
			}
		}

# --restart
		elsif ( $ARGV[0] =~ /^((--)??restart|-R)$/ ) {
			shift @ARGV if ( defined $ARGV[1] and $playd_novid = $ARGV[1] =~ /^(--)??novid$/ );
			shift @ARGV if ( defined $ARGV[1] and $playd_console = $ARGV[1] =~ /^(--)??console$/ );
			&quit;
			&start;
		}

# --seek
		elsif ( $ARGV[0] =~ /^((--)??seek|-s)$/ ) {
			if ( defined $ARGV[1] and $ARGV[1] =~ /^[+-]??(\d+|\0)(m|h)??$/ ) {
				( my $seek_time ) = $ARGV[1] =~ /^(\d+)/;
				$seek_time *= 60 if ( $ARGV[1] =~ /m$/ );
				$seek_time *= 3600 if ( $ARGV[1] =~ /h$/ );

				if ( defined $ARGV[2] and $ARGV[2] =~ /^(--)??abs(olute)??$/ ) {
					&send_mplayer_cmd("seek $seek_time 2");
					shift @ARGV;
					shift @ARGV;
				}
				elsif ( defined $ARGV[2] and $ARGV[2] =~ /^(--)??percent|%$/ ) {
					unless ( $ARGV[1] =~ /(m|h)$/ ) {
						&send_mplayer_cmd("seek $ARGV[1] 1");
						shift @ARGV;
						shift @ARGV;
					}
					else {
						warn "WARN: $ARGV[2] doesn't support 'm' and 'h' modifiers. Ignoring\n";
						shift @ARGV;
						shift @ARGV;
					}
				}
				else {
					&send_mplayer_cmd("seek $seek_time 0");
					shift @ARGV
				}
			}
			else {
				warn "WARN: $ARGV[0] needs at least numerica argument. Ignoring\n";
			}
		}

		elsif ( $ARGV[0] =~ /^((--)??start|-S)$/ ) {
			shift @ARGV if ( defined $ARGV[1] and $playd_novid = $ARGV[1] =~ /^(--)??novid$/ );
			shift @ARGV if ( defined $ARGV[1] and $playd_console = $ARGV[1] =~ /^(--)??console$/ );
			&start;
		}

# --volume
# --brightness
# --contrast
# --gamma
# --hue
# --saturation
# --audio-delay
# --loop
		elsif ( $ARGV[0] =~ /^((--)??(vol(ume)??|brightness|contrast|gamma|hue|saturation|audio-delay|loop)|-v)$/ ) {
			my $parameter = $ARGV[0];
			$parameter =~ tr/-//;
			$parameter = 'audio_delay' if ( $parameter =~ /audiodelay/ );
			$parameter = 'volume' if ( $parameter =~ /vol(ume)??|V/ );

			if ( defined $ARGV[1] and $ARGV[1] =~ /^[+-]??\d+$/ ) {
				if ( defined $ARGV[2] and $ARGV[2] =~ /^(--)??abs(olute)??$/ ) {
					&send_mplayer_cmd("$parameter $ARGV[1] 1");
					shift @ARGV;
					shift @ARGV;
				}
				else {
					&send_mplayer_cmd("$parameter $ARGV[1] 0");
					shift @ARGV
				}
			}
			else {
				warn "WARN: $ARGV[0] needs at least numerica argument. Ignoring\n";
			}
		}

# --next
		elsif ( $ARGV[0] =~ /^((--)??next|-n)$/ ) {
			&send_mplayer_cmd('seek 100 1');
		}

# --version
		elsif ( $ARGV[0] =~ /^(--)??version$/ ) {
			print "$playd_name v$playd_version\n";
		}

# --switch-audio
		elsif ( $ARGV[0] =~ /^(--)??sw(itch)??-audio$/ ) {
			&send_mplayer_cmd('switch_audio');
		}

# --switch-subtitle
		elsif ( $ARGV[0] =~ /^(--)??(switch-substitle|sw-sub)$/ ) {
			&send_mplayer_cmd('sub_select');
		}

# --cmd
		elsif ( $ARGV[0] =~ /^(--)??cmd$/ ) {
			if ( defined $ARGV[1] ) {
				&send_mplayer_cmd($ARGV[1]);
				shift @ARGV;
			}
			else {
				warn "WARN: $ARGV[0] need option to be passed to mplayer. Ignoring\n";
				warn "WANR: read http://www.mplayerhq.hu/DOCS/tech/slave.txt\n";
			}
		}

# --pause
		elsif ( $ARGV[0] =~ /^((--)??pause|-z)$/ ) {
			&send_mplayer_cmd('pause');
		}

# --again
		elsif ( $ARGV[0] =~ /^((--)??again|-A)$/ ) {
			&send_mplayer_cmd('seek 0 1');
		}

# --noscheck
		elsif ( $ARGV[0] =~ /^(--)??nocheck$/ ) {
			if ( defined $ARGV[1] and -f $ARGV[1] ) {
				&add_to_playlist($ARGV[1]);
				shift @ARGV;
			}
			else {
				warn "WARN: '$ARGV[1]' is not a file. Ignoring\n";
			}
		}

# STREAMS
		elsif ( $ARGV[0] =~ /:\/\// ) {
			&start;
			&send_mplayer_cmd("loadfile $ARGV[0] $playd_append");
		}

# --file
# ANYTHING ELSE
		else {
			shift @ARGV if ( defined $ARGV[1] and $ARGV[0] =~ /^((--)??file|-f)$/ );
			my $src = abs_path($ARGV[0]);
			if ( -f $src ) {
				my @tmp = &playable($src);
				&add_to_playlist($tmp[0]) if ( $tmp[0] );
			}
			elsif ( -d $src ) {
				&start;
				&mk_playlist($src);
				if ( $playd_append ) {
					open(TMP_PLAYLIST, '<', $playd_tmp_playlist) or die "ERR: can't open file '$playd_tmp_playlist' for reading.\n$!\n";
					open(PLAYLIST, '>>', $playd_playlist) or die "ERR: can't open file '$playd_playlist' for writing.\n$!\n";
					print PLAYLIST <TMP_PLAYLIST>;
					close TMP_PLAYLIST;
					close PLAYLIST;

					&send_mplayer_cmd("loadlist '$playd_tmp_playlist' 1");
					unlink $playd_tmp_playlist or die "ERR: can't delete file '$playd_tmp_playlist'.\n$!\n";
				}
				else {
					unlink $playd_playlist if ( -f $playd_playlist );
					rename ($playd_tmp_playlist, $playd_playlist) or die "ERR: $!\n";
					&send_mplayer_cmd("loadlist '$playd_playlist' 0");
				}
				$playd_append = 1;

			}
			else {
				warn "WARN: '$ARGV[0]' not a file/directory. Ignoring\n";
			}
		}

		shift @ARGV;
	}
	exit;
}

sub send_mplayer_cmd {
##############################
# Send command to mplayer pipe
##############################
	if ( &get_mplayer_pid ) {
		open (PIPE, '>>', $playd_fifo) or die "ERR: can't open '$playd_fifo':\n$!\n";
		foreach my $item (@_) {
			print PIPE "$item\n";
			sleep 1;
		}
		close PIPE;
	}
	else { die "ERR: playd not started. Run:\n$playd_name start\n"; }
}

sub add_to_playlist {
##############################
# add items to playlist
# nothing will be checked, check should happen earlier
##############################
	&start;
	unless ( $playd_append ) {
		unlink $playd_playlist if ( -f $playd_playlist ); 
	}
	open (PLAYLIST, '>>', $playd_playlist) or die "ERR: can't open '$playd_playlist':\n$!\n";

	foreach my $item (@_) {
		&send_mplayer_cmd("loadfile '$item' $playd_append");
		print PLAYLIST "$item\n";
		$playd_append = 1;
	}
	close PLAYLIST;
}

sub help () {
##############################
# desplays help
##############################
	print << "EOF"

  $playd_name\tv$playd_version
  by:		killasmurf86
  e-mail:	killasmurf86\@gmail.com
  homepage: http://code.google.com/p/playd/

  $playd_name v$playd_version is a rewrite of playd.sh Bourne Shell script in perl.
  It controls mplayer daemon via pipe. It's easy to use and implement in
  other apps (In FVWM2 menu for example).

  I wrote playd.sh when I was learning Bourne Shell. Now I'm learning Perl and
  I think this is good script to practice what I read in book.

  Hopefully this version will be more portable than playd.sh version :D

EOF
}

sub get_mplayer_pid () {
##############################
# chech if playd is running, remove lock if not
# returns GID if playd is running, return 0 if not or GIDS don't match
##############################
	if ( -f $playd_lock ) {
		open (PIDFILE, '<', $playd_lock) or return 0;
		my $gid = <PIDFILE>;
		chomp $gid;
		close PIDFILE;

		my $pid = `pgrep -g $gid -n mplayer`;
		chomp $pid;
		return $pid if ( $pid );

		unlink $playd_lock if ( -f $playd_lock ); 
	}
	return 0;
}

sub list {
##############################
# lists current playlist
# if $_[0] > 0 , then list files including full path
# otherwhise $_[0] must be 0
##############################
	open (PLAYLIST, '<', $playd_playlist) or die "ERR: can't open file '$playd_playlist' for reading.\n$!\n";
	my $i = 1;
	while (my $playlist_item = <PLAYLIST>) {
		# if nessacery remove file path
		$playlist_item =~ s/.*?\///g unless ( $_[0] );

		print "$i) $playlist_item";
		$i++;
	}
	close PLAYLIST;
}

sub start () {
##############################
# start mplayer in slave mode
##############################
	unless ( &get_mplayer_pid ) {
		my $mplayer_cmd;
		unless ( $playd_novid ) {
			$mplayer_cmd = "$playd_mplayer_cmd $playd_mplayer_user_options";
		}
		else {
			$mplayer_cmd = "$playd_mplayer_no_video_cmd $playd_mplayer_user_options";
		}

		open(PIDFILE, '>', $playd_lock) or die "ERR: can't open '$playd_lock' for writing.\n$!\n";
		print PIDFILE "$$\n";
		close PIDFILE;

		my $dir = getcwd;
		chdir rootdir;

		unless ( $playd_console ) {
			system("$mplayer_cmd > /dev/null 2> /dev/null &");
			sleep 1;
		}
		else {
			my $pid;
			FORK: {
				if ( $pid = fork ) {
					chdir $dir;
					return;
				}
				elsif ( defined $pid ) {
					system($mplayer_cmd);
					unlink $playd_lock if ( -f $playd_lock );
					exit;
				}
				elsif ( $! == EAGAIN ) {
					sleep 5;
					redo FORK;
				}
				else {
					unlink $playd_lock if ( -f $playd_lock );
					die "ERR: failed to fork\n";
				}
			}
		}

		chdir $dir;
	}
}

sub quit () {
##############################
# quit mplayer in slave mode
##############################
	my $pid = &get_mplayer_pid;
	if ( $pid ) {
		&send_mplayer_cmd('quit');
		sleep 1;
		if ( kill(0, $pid) ) {
			kill 3, -$pid;
			sleep 1;
			if ( kill(0, $pid) ) {
				kill 9, -$pid;
				sleep 1;
				die "ERR: can't kill mplayer. PID: $pid\n" unless ( kill(0, $pid) );
			}
		}
		unlink $playd_lock if ( -f $playd_lock );
	}
}

sub mk_playlist {
##############################
# make playlist form all files in dir and subdirs
# arg[0 .. n] = absolute path to files/dirs
##############################
	my @file_list;
	my @new_file_list;
	foreach my $item (@_) {
		if ( -f $item ) {
			push(@file_list, "$item\n");
		} elsif ( -d $item ) {
			opendir(CURRENT_DIR, $item) or die "ERR: can't read dircetory: '$item'.\n$1\n";
			my @content = readdir CURRENT_DIR;
			closedir(CURRENT_DIR);
			foreach my $i (2 .. $#content) {
				$content[$i] = $item . "/" . $content[$i];
				push(@new_file_list, $content[$i]);
			}
		}
	}

	open (LIST, '>>', $playd_tmp_playlist) or die "ERR: can't open file '$playd_tmp_playlist' for writing.\n$!\n";
	print LIST &playable(sort @file_list);
	close LIST;
	&mk_playlist(@new_file_list) if ( @new_file_list );
}

sub playable () {
##############################
# return arry with playable files
# input arry of file name or scalar with file name
# file names, may end with \n
##############################
	my @playable_list;
	foreach my $item (@_) {
		push(@playable_list, $item)
			if ( $item =~ /\.(mp3|flac|og[ga]|aac|wav|wma|aif[cf]??|m4[abpr]|mp[21a]|mpga|mid[i]??|mka|avi|mkv|mpf|mp4|wmv|vob|3gp|mov|d[i]??vx|og[mv])(\n)??$/i
			or $ft->checktype_filename($item) =~ /^(audio|video)/ );
	}
	return @playable_list;
}

